"
 COPYRIGHT (c) 1989 by Claus Gittinger
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"

View subclass:#InspectorView
       instanceVariableNames:'listView workspace 
			      inspectedObject selectedLine
			      inspectedValues nShown menu1 menu2'
       classVariableNames:''
       poolDictionaries:''
       category:'Interface-Inspector'
!

InspectorView comment:'
COPYRIGHT (c) 1989 by Claus Gittinger
	      All Rights Reserved

$Header: /cvs/stx/stx/libtool/Attic/InspView.st,v 1.12 1994-11-17 14:46:51 claus Exp $
'!

!InspectorView class methodsFor:'documentation'!

copyright
"
 COPYRIGHT (c) 1989 by Claus Gittinger
	      All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"
!

version
"
$Header: /cvs/stx/stx/libtool/Attic/InspView.st,v 1.12 1994-11-17 14:46:51 claus Exp $
"
!

documentation
"
    This class implements an graphical inspector.
    Inspecting can be done on an object -
    (where its instvarnames/values are inspected)
    or a list of objects (where a namearray/valuesarray is inspected).
    The later is used by the debugger to inspect method variables/args.

    The system calls the inspector through the global variable 'Inspector'
    which is bound to this class (but could be redefined - it actually is
    set to MiniInspector in a smalltalk without graphical user interface).

    Also notice, that there are two methods to inspect an object:
    sending #basicInspect to any object, will open this kind of inspector on 
    it (showing instance variables which are physically present).

    Sending it #inspect, will - depending on the object - sometimes invoke a
    specialized inspector. 
    (see OrderedCollectionInspectorView, ImageInspectorView etc. as examples).

    examples:
	    #(1 2 3 4) asOrderedCollection inspect
	    #(1 2 3 4) asOrderedCollection basicInspect
	    (Array new:10000) inspect
	    (Image fromFile:'bitmaps/claus.gif') inspect
	    (Image fromFile:'bitmaps/claus.gif') basicInspect
	    (Image fromFile:'bitmaps/SBrowser.xbm') inspect
	    (Image fromFile:'bitmaps/SBrowser.xbm') basicInspect
"
! !

!InspectorView class methodsFor:'instance creation'!

for:anObject
    "create and launch a new inspector for anObject.
     This protocol is a historic leftover - this method will vanish."

    ^ self openOn:anObject
!

inspect:anObject
    "create and launch a new inspector for anObject.
     This protocol is a historic leftover - this method will vanish."

    ^ self openOn:anObject
!

openOn:anObject
    "create and launch a new inspector for anObject"

    |topView inspectorView nm|

    anObject isClass ifTrue:[
	nm := anObject displayString
    ].
    nm isNil ifTrue:[
	nm := anObject classNameWithArticle
    ].

    topView := StandardSystemView
		    label:('Inspector on: ' , nm)
		     icon:(Form fromFile:'Inspector.xbm' resolution:100)
		minExtent:(100 @ 100).

    topView extent:(Display width // 3) @ (Display height // 3).

    inspectorView := self origin:(0.0 @ 0.0)
			  corner:(1.0 @ 1.0)
			     in:topView.

    "kludge: must realize first, to be able to set menu again"
    topView open.
    inspectorView inspect:anObject.

    "
     InspectorView openOn:(5 @ 7)
     InspectorView openOn:(Array new:400)
     DictionaryInspectorView openOn:(IdentityDictionary new)
    "
! !

!InspectorView methodsFor:'initialization'!

initialize
    |v panel|

    super initialize.

    panel := VariableHorizontalPanel origin:(0.0 @ 0.0)
				     corner:(1.0 @ 1.0)
					 in:self.

    v := HVScrollableView for:SelectionInListView 
			  miniScrollerH:true
			  miniScrollerV:false
			  in:panel.
    v origin:(0.0 @ 0.0) corner:(0.3 @ 1.0).
    listView := v scrolledView.
    listView action:[:lineNr | self showSelection:lineNr].
    listView doubleClickAction:[:lineNr | self doInspect].
    listView ignoreReselect:false.

    v := HVScrollableView for:CodeView 
			  miniScrollerH:true
			  miniScrollerV:false
			  in:panel.
    v origin:(0.3 @ 0.0) corner:(1.0 @ 1.0).
    workspace := v scrolledView.

    workspace acceptAction:[:theText | self doAccept:theText asString].

    nShown := 100.
    self initializeListViewMiddleButtonMenus
!

initializeListViewMiddleButtonMenus
    menu1 := PopUpMenu
		  labels:(resources array:#(
					    'inspect'
"/                                          '-'
"/                                          'owners'
					   ))
	       selectors:#(
			   doInspect
"/                           nil
"/                           inspectOwners
			  )
		receiver:self
		     for:listView.
    menu2 := PopUpMenu
		  labels:(resources array:#(
					    'inspect'
					    '-'
					    'show more'
					   ))
	       selectors:#(doInspect nil showMore)
		receiver:self
		     for:listView.
!

mapped
    "delayed setup of lists till map-time - 
     this makes startup of inspectors a bit faster"

    |o|

    super mapped.
    "
     kludge to trick inspect:, which ignores setting the
     same object again ...
    "
    o := inspectedObject.
    inspectedObject := nil.
    self inspect:o
! !

!InspectorView methodsFor:'release'!

release
    "release inpected object. This is normally not needed,
     since the garbage collector will find this memory alone.
     However, if some applications keeps inspectors (for example,
     the debugger does this), this would be freed very late."

"
    inspectedObject notNil ifTrue:[
	inspectedObject removeDependent:self
    ].
"
    inspectedObject := nil.
    inspectedValues := nil.
    workspace doItAction:nil.
    workspace contents:nil.
    listView contents:nil
! !

!InspectorView methodsFor:'private'!

listOfNames
    "return a list of names to show in the selectionList"

    |aList n cut cls|

    aList := OrderedCollection new.
    aList add:'self'.
    cls := inspectedObject class.
    cls isClass ifTrue:[
	cls allInstVarNames do:[:instVarName |
	    aList add:instVarName
	]
    ] ifFalse:[
	1 to:cls instSize do:[:index |
	    aList add:('instvar' , index printString)
	]
    ].
    cls isVariable ifTrue:[
	n := inspectedObject basicSize.
	(n > nShown) ifTrue:[
	    n := nShown.
	    cut := true.
	    listView setMiddleButtonMenu:menu2.
	] ifFalse:[
	    cut := false.
	    listView setMiddleButtonMenu:menu1.
	].
	1 to:n do:[:index |
	    aList add:(index printString)
	].
	cut ifTrue:[
	    aList add:' ... '
	]
    ].
    ^ aList
!

showMore
    nShown := nShown * 2.
    self inspect:inspectedObject
!

setDoItAction
    workspace doItAction:[:theCode |
	(inspectedObject class compiler) 
	    evaluate:theCode 
	    in:nil 
	    receiver:inspectedObject 
	    notifying:workspace 
	    logged:true 
	    ifFail:nil 
    ].

! !

!InspectorView methodsFor:'accessing'!

inspect:anObject
    "define the object to be inspected"

    |aList sameObject|

    sameObject := anObject == inspectedObject.
"
    sameObject ifFalse:[
	inspectedObject notNil ifTrue:[
	    inspectedObject removeDependent:self
	]
    ].
"
    inspectedObject := anObject.
    self initializeListViewMiddleButtonMenus.
    listView setMiddleButtonMenu:menu1.
    realized ifFalse:[^ self].

    aList := self listOfNames.
    sameObject ifTrue:[
	listView setContents:aList
    ] ifFalse:[
	listView contents:aList
    ].

    workspace contents:nil.
    self setDoItAction.

"
    sameObject ifFalse:[
	inspectedObject notNil ifTrue:[
	    inspectedObject addDependent:self
	]
    ].
"
    inspectedValues := nil.
    selectedLine := nil
!

inspect:anObject values:valueArray names:nameArray
    listView contents:nameArray.
    workspace contents:nil.
"
    inspectedObject notNil ifTrue:[
	inspectedObject removeDependent:self
    ].
"
    inspectedObject := anObject.
"
    inspectedObject notNil ifTrue:[
	inspectedObject addDependent:self
    ].
"
    inspectedValues := valueArray.
    selectedLine := nil
!

inspectValues:valueArray names:nameArray
    listView contents:nameArray.
    workspace contents:nil.
"
    inspectedObject notNil ifTrue:[
	inspectedObject removeDependent:self
    ].
"
    inspectedObject := nil.
    inspectedValues := valueArray.
    selectedLine := nil
! !

!InspectorView methodsFor:'user interaction'!

keyPress:aKey x:x y:y
    "all my input is passed on to the workspace-field"

    workspace keyPress:aKey x:0 y:0
!

update:something with:someArgument from:changedObject
    "handle updates from other inspectors"

    |oldSelection|

    changedObject == inspectedObject ifTrue:[
	oldSelection := listView selection.
	self inspect:inspectedObject.
	oldSelection notNil ifTrue:[
	    self showSelection:oldSelection
	]
    ]
!

destroy
    inspectedObject notNil ifTrue:[
"
	inspectedObject removeDependent:self.
"
	inspectedObject := nil
    ].
    menu1 notNil ifTrue:[
	menu1 destroy. menu1 := nil.
    ].
    menu2 notNil ifTrue:[
	menu2 destroy. menu2 := nil.
    ].
    inspectedValues := nil.
    super destroy
!

showSelection:lineNr
    "user clicked on an instvar - show value in workspace"

    |val string index|

"
    workspace contents:nil.
"
    (lineNr == listView list size
     and:[(listView listAt:lineNr) startsWith:' ...']) ifTrue:[
	"clicked on the '...' entry"
	self showMore.
	listView selection:lineNr.
    ].

    inspectedValues isNil ifTrue:[
	lineNr == 1 ifTrue:[
	    "selecting self also does a re-set, this allows updating the list"
	    self inspect:inspectedObject.
	    val := inspectedObject
	] ifFalse:[
	    index := lineNr - 1.
	    (inspectedObject class isVariable) ifFalse:[
		val := inspectedObject instVarAt:index
	    ] ifTrue:[
		index <= (inspectedObject class instSize) ifTrue:[
		    val := inspectedObject instVarAt:index
		] ifFalse:[
		    index := index - inspectedObject class instSize.
		    val := inspectedObject basicAt:index
		]
	    ]
	]
    ] ifFalse:[
	val := inspectedValues at:lineNr
    ].
    string := val displayString.
"
    workspace cursorToTop.
"
    workspace replace:string.

    selectedLine := lineNr
!

doAccept:theText
    |value index|

    value := inspectedObject class compiler 
		 evaluate:theText
		 receiver:inspectedObject 
		 notifying:workspace.
    inspectedValues isNil ifTrue:[
	selectedLine notNil ifTrue:[
	    selectedLine == 1 ifFalse:[
		index := selectedLine - 1.
		(inspectedObject class isVariable) ifFalse:[
		    inspectedObject instVarAt:index put:value
		] ifTrue:[
		    index <= (inspectedObject class instSize) ifTrue:[
			inspectedObject instVarAt:index put:value
		    ] ifFalse:[
			index := index - inspectedObject class instSize.
			inspectedObject basicAt:index put:value
		    ]
		]
	    ]
	]
    ] ifFalse:[
	selectedLine notNil ifTrue:[
	    inspectedValues at:selectedLine put:value
	]
    ].
    inspectedObject changed
!

doInspect
    "user selected inspect-menu entry"
    |index objectToInspect|

    selectedLine notNil ifTrue:[
	inspectedValues isNil ifTrue:[
	    (selectedLine == 1) ifTrue:[
		objectToInspect := inspectedObject
	    ] ifFalse:[
		index := selectedLine - 1.
		(inspectedObject class isVariable) ifFalse:[
		    objectToInspect := inspectedObject instVarAt:index
		] ifTrue:[
		    index <= (inspectedObject class instSize) ifTrue:[
			objectToInspect := inspectedObject instVarAt:index
		    ] ifFalse:[
			index := index - inspectedObject class instSize.
			objectToInspect := inspectedObject basicAt:index
		    ]
		]
	    ]
	] ifFalse:[
	    objectToInspect := inspectedValues at:selectedLine ifAbsent:[^ self]
	].
	objectToInspect inspect
    ]
!

inspectOwners
    self withCursor:(Cursor questionMark) do:[
	|owners dict|

	owners := (ObjectMemory whoReferences:inspectedObject) asOrderedCollection.
	owners size > 500 ifTrue:[
	    (self confirm:'there are ' , owners size printString , ' owners.\\Do you really want to see them all ?' withCRs)
	    ifFalse:[^ self]
	].
	dict := IdentityDictionary new.
	owners do:[:owner |
	    |set names oClass|

	    owner ~~ self ifTrue:[
		set := Set new.
		names := owner class allInstVarNames.
		oClass := owner class.
		1 to:oClass instSize do:[:i |
		    (owner instVarAt:i) == inspectedObject ifTrue:[
			set add:(names at:i).
		    ].
		].
		oClass isVariable ifTrue:[
		    oClass isPointers ifTrue:[
			1 to:owner basicSize do:[:i |
			    (owner basicAt:i) == inspectedObject ifTrue:[
				 set add:i
			    ]
			]
		    ]
		].
		dict at:owner put:set
	    ].
	].
	dict inspect
    ]
! !
